import numpy as np
import scipy as sp


class TiseSolverFgh1D:
    def __init__(
        self,
        xa: float,
        xb: float,
        nx: int,
    ):
        x = np.linspace(xa, xb, nx, endpoint=False, dtype=np.float64)
        self.xa: float = xa
        self.xb: float = xb
        self.nx: int = nx
        self.x: np.ndarray = x
        self.Vx: np.ndarray = self.V(x)

    # TISE potential
    def V(self, x: np.array):
        V = 0.5 * x * x
        return V

    def abscissa_potential(self):
        return self.x, self.Vx

    def solve(self):
        # Hamiltonian matrix (N x N)
        matH = TiseSolverFgh1D.matH(self.x, self.Vx)

        # Diagonalize
        eigenvalues, eigenvectors = sp.linalg.eig(matH, left=False, right=True)

        # Sort eigenvalues and eigenvectors in assending order by real part of the eigenvalues.
        idx = np.argsort(eigenvalues.real)
        eigenvalues = eigenvalues[idx]
        eigenvectors = eigenvectors[:, idx]

        # Normalize to \int |\psi(x)|^2 dx = 1, (Just in case)
        S = np.sum(np.abs(eigenvectors) ** 2, axis=0)
        eigenvectors /= np.sqrt(S)

        # Eigenvector to eigenfunction
        eigenfunctions = eigenvectors.astype(np.float64) * np.sqrt(self.nx / (self.xb - self.xa))

        return eigenvalues.real, eigenfunctions

    @staticmethod
    def matH(
        x: np.ndarray,
        V: np.ndarray,
    ):
        N = x.shape[0]  # = V.shape[0]

        # Initialize the Hamiltonian matrix
        ham = np.zeros((N, N), dtype=np.float64)

        # Calculate grid spacing and wave vector
        dx = x[1] - x[0]
        K = 2.0 * np.pi / dx

        is_code_optimization: bool = True
        if is_code_optimization:
            # Precompute indices differences and sine/cosine values
            indices = np.arange(N)
            i_minus_j = indices[:, None] - indices[None, :]
            sin_term = np.sin(np.pi * i_minus_j / N)

            # Handle diagonal elements separately
            diagonal_factor = (N**2 + 2) / 12.0 if N % 2 == 0 else (N**2 - 1) / 12.0
            diagonal = 0.5 * diagonal_factor * (K / N) ** 2 + V

            # Compute off-diagonal elements using broadcasting
            off_diagonal = 0.5 * (-1) ** np.abs(i_minus_j)
            mask = i_minus_j != 0  # Only off_diagonal components to avoid zero-division
            if N % 2 == 0:
                denom = sin_term**2
                denom[~mask] = 1.0  # Replace non-zero value to avoid zero-division
                off_diagonal = off_diagonal / denom
            else:
                cos_term = np.cos(np.pi * i_minus_j / N)
                denom = sin_term**2
                denom[~mask] = 1.0
                off_diagonal = off_diagonal * cos_term / denom

            # Eliminate infinities or NaNs caused by division by zero on diagonal
            np.fill_diagonal(off_diagonal, 0)

            # Scale the off-diagonal terms
            off_diagonal *= 0.5 * (K / N) ** 2

            # Combine diagonal and off-diagonal components
            np.fill_diagonal(ham, diagonal)
            ham += off_diagonal

        else:
            # Legacy code (not optimized, 20~50 times slower than optimized code)
            if N % 2 == 0:
                for i in range(N):
                    for j in range(N):
                        if j == i:
                            ham[i, j] = 0.5 * ((N**2 + 2) / 12.0) * (K / N) ** 2 + V[i]
                        else:
                            ham[i, j] = 0.5 * (0.5 * (-1) ** (i - j) / ((np.sin(np.pi * (i - j) / N)) ** 2)) * (K / N) ** 2
            else:
                for i in range(N):
                    for j in range(N):
                        if j == i:
                            ham[i, j] = 0.5 * ((N**2 - 1) / 12.0) * (K / N) ** 2 + V[i]
                        else:
                            ham[i, j] = (
                                0.5 * (0.5 * (-1) ** (i - j) * np.cos(np.pi * (i - j) / N) / ((np.sin(np.pi * (i - j) / N)) ** 2)) * (K / N) ** 2
                            )

        return ham

    def fgh_basis(
        self,
        n: int,
        x: np.ndarray,
    ):
        """
        Call script example
        # -- Output TISE results
        new_x = np.linspace(xa, xb, 10 * nx + 1, endpoint=True, dtype=np.float64)
        x_dict = {}
        y_dict = {}
        for n in range(5):
            x_dict[f"Idx {n:>2d}"] = np.array(new_x)
            y_dict[f"Idx {n:>2d}"] = solver.fgh_basis(n, new_x)
        plot_xy(x_dict, y_dict, show_gui=True)
        """
        L = self.xb - self.xa
        if self.nx % 2 == 0:
            l_idx = np.arange(-self.nx // 2, self.nx // 2)  # l ranges from -N/2 to N/2-1
        else:
            l_idx = np.arange(-(self.nx - 1) // 2, (self.nx - 1) // 2 + 1)  # l ranges from -(N-1)/2 to (N-1)/2

        # print(l_idx.shape[0])
        exponent = 2j * np.pi * l_idx[:, None] * ((x[None, :] - self.xa) / L - n / self.nx)
        summation = np.sum(np.exp(exponent), axis=0)  # sum each column
        varphi = summation / np.sqrt(self.nx * L)

        return varphi
